"""Loop graph constructor for GNNs."""

import numpy as np

from .base import GraphBase

__all__ = ["LoopGraph"]


class LoopGraph(GraphBase):
    """Generates loop-only graphs.

    Connects every node in the graph with itself but nothing else.

    See :class:`GraphBase` for attributes/methods shared
    across all graph constructors.
    """

    # Name of the graph constructor (as specified in the configuration)
    name = "loop"

    def generate(self, clusts, **kwargs):
        """Generate a loop-graph on a set of N nodes.

        Parameters
        ----------
        clusts : IndexBatch
            (C) Cluster indexes
        **kwargs : dict, optional
            Unused graph generation arguments

        Returns
        -------
        np.ndarray
            (2, E) Tensor of edges
        np.ndarray
            (B) Number of edges in each entry of the batch
        """
        # There is exactly one edge per cluster
        edge_counts = clusts.counts

        # Define the loop graph
        num_nodes = np.sum(edge_counts)
        edge_index = np.repeat(np.arange(num_nodes)[None, :], 2, axis=0)

        return edge_index, edge_counts



"""Module which defines a graph node feature update based on AGNNConv."""

from torch import nn
from torch_geometric.nn import AGNNConv

from spine.model.layer.common.act_norm import act_factory, norm_factory

__all__ = ["AGNNConvNodeLayer"]


class AGNNConvNodeLayer(nn.Module):
    """AGNNConv module for extracting graph node features.

    This model simply takes a simple attention-based convolution of a node
    with all of its neighbors to update the initial node feature vector (N_c),
    preserving the original size and returning an updated (N_c) vector.

    Source: https://arxiv.org/abs/1803.03735
    """

    # Name of the node layer (as specified in the configuration)
    name = "agnnconv"

    def __init__(self, node_in, edge_in, glob_in, activation, normalization, **kwargs):
        """Initialize the MLPs which are used to update the node features.

        Parameters
        ----------
        node_in : int
            Number of input node features
        edge_in : int
            Number of input edge features
        glob_in : int
            Number of input global features for the graph
        activation : Union[str, dict], default 'relu'
            Activation function configuration
        normalization : Union[str, dict], default 'batch_norm'
            Normalization function configuration
        **kwargs : dict
            Extra parameters to be passed to the AGNNConv layer
        """
        # Initialize the parent class
        super().__init__()

        # Initialize the layer
        self.feature_size = node_in
        self.agnnconv = AGNNConv(**kwargs)
        self.bn = norm_factory(normalization, node_in)
        self.act = act_factory(activation)

    def forward(self, node_feats, edge_index, *args):
        """Pass a batch of node/edges through the edge update layer.

        Parameters
        ----------
        node_feats : torch.Tensor
            (C, N_c) Node features
        edge_index : torch.Tensor
            (2, E) Incidence matrix
        *args : list, optional
            Other parameters passed but not needed

        Returns
        -------
        torch.Tensor
            (C, N_o) Updated node features
        """
        x = self.agnnconv(node_feats, edge_index)
        x = self.bn(x)
        x = self.act(x)

        return x



"""Algorithms associated with the deghosting process."""

import numpy as np
import torch

from spine.data import TensorBatch

from .globals import DELTA_SHP, MICHL_SHP, SHOWR_SHP, TRACK_SHP


class ChargeRescaler:
    """Rescales the space point charge based on the deghosting output.

    It ensures that the amount of charge carried by each hit that makes up at
    least one space point is not duplicated by distributing said hit charge
    across all the space points formed with it.
    """

    def __init__(self, collection_only=False, collection_id=2):
        """Initialize the charge rescaler.

        Parameters
        ----------
        collection_only : bool, default False
            If `True`, only use the collection plane to estimate the rescaled charge
        collection_id : int, default 2
            Index of the collection plane
        """
        # Save the parameters
        self.collection_only = collection_only
        self.collection_id = collection_id

    def __call__(self, data):
        """Rescale the charge of one batch of deghosted data.

        Parameters
        ----------
        data : TensorBatch
            (N, 1 + D + N_f + 6) tensor of voxel/value pairs

        Returns
        -------
        data : Union[np.ndarray, torch.Tensor]
            (N) Rescaled charge values
        """
        charges = data._empty(len(data.tensor))
        for b in range(data.batch_size):
            lower, upper = data.edges[b], data.edges[b + 1]
            charges[lower:upper] = self.process_single(data[b])

        return charges

    def process_single(self, data):
        """Rescale the charge of one event.

        The last 6 columns of the input tensor *MUST* contain:
        - charge in each of the projection planes (3)
        - unique index of the hit in each 2D projection (3)

        Notes
        -----
        This function should work on numpy arrays or Torch tensors.

        Parameters
        ----------
        data : Union[np.ndarray, torch.Tensor]
            (N, 1 + D + N_f + 6) tensor of voxel/value pairs

        Returns
        -------
        data : Union[np.ndarray, torch.Tensor]
            (N) Rescaled charge values
        """
        # Define operations on the basis of the input type
        if torch.is_tensor(data):
            unique, where = torch.unique, torch.where
            sum = lambda x: torch.sum(x, dim=1)
        else:
            unique, where = np.unique, np.where
            sum = lambda x: np.sum(x, axis=1)

        # Count how many times each wire hit is used to form a space point
        hit_ids = data[:, -3:]
        _, inverse, counts = unique(hit_ids, return_inverse=True, return_counts=True)
        mult = counts[inverse].reshape(-1, 3)

        # Rescale the charge on the basis of hit multiplicity
        hit_charges = data[:, -6:-3]
        if not self.collection_only:
            # Take the average of the charge estimates from each active plane
            pmask = hit_ids > -1
            charges = sum((hit_charges * pmask) / mult) / sum(pmask)
        else:
            # Only use the collection plane measurement, when available
            charges = hit_charges[:, self.collection_id] / mult[:, self.collection_id]

            # Fallback on the average if there is no collection hit
            bad_index = where(hit_ids[:, self.collection_id] < 0)[0]
            if len(bad_index) > 0:
                pmask = hit_ids[bad_index] > -1
                charges[bad_index] = sum(
                    (hit_charges[bad_index] * pmask) / mult[bad_index]
                ) / sum(pmask)

        return charges



"""Module with a data class object which represents trigger information.

This copies the internal structure of :class:`larcv.Trigger`.
"""

from dataclasses import dataclass

from .base import DataBase

__all__ = ["Trigger"]


@dataclass(eq=False)
class Trigger(DataBase):
    """Trigger information.

    Attributes
    ----------
    id : int
        Trigger ID
    time_s : int
        Integer seconds component of the UNIX trigger time
    time_ns : int
        Integer nanoseconds component of the UNIX trigger time
    beam_time_s : int
        Integer seconds component of the UNIX beam pulse time
    beam_time_ns : int
        Integer seconds component of the UNIX beam pulse time
    type : int
        DAQ-specific trigger type
    """

    id: int = -1
    time_s: int = -1
    time_ns: int = -1
    beam_time_s: int = -1
    beam_time_ns: int = -1
    type: int = -1

    @classmethod
    def from_larcv(cls, trigger):
        """Builds and returns a Trigger object from a LArCV Trigger object.

        Parameters
        ----------
        trigger : larcv.Trigger
            LArCV-format trigger information

        Returns
        -------
        Trigger
            Trigger object
        """
        return cls(
            id=trigger.id(),
            time_s=trigger.time_s(),
            time_ns=trigger.time_ns(),
            type=trigger.type(),
        )



"""Module with a dataclass targeted at batched matrix/tensors."""

from dataclasses import dataclass

import numpy as np

from spine.utils.conditional import ME, torch
from spine.utils.docstring import inherit_docstring
from spine.utils.globals import BATCH_COL, COORD_COLS

from .base import BatchBase

__all__ = ["TensorBatch"]


@dataclass(eq=False)
@inherit_docstring(BatchBase)
class TensorBatch(BatchBase):
    """Batched tensor with the necessary methods to slice it."""

    def __init__(
        self,
        data,
        counts=None,
        batch_size=None,
        is_sparse=False,
        has_batch_col=False,
        coord_cols=None,
    ):
        """Initialize the attributes of the class.

        Parameters
        ----------
        data : Union[np.ndarray, torch.Tensor, ME.SparseTensor]
            (N, C) Batched tensors
        counts : Union[List[int], np.ndarray, torch.Tensor]
            (B) Number of data rows in each entry
        batch_size : int, optional
            Number of entries that make up the batched data
        is_sparse : bool, default False
            If initializing from an ME sparse data, flip to True
        has_batch_col : bool, default False
            Wheather the tensor has a column specifying the batch ID
        coord_cols : Union[List[int], np.ndarray], optional
            List of columns specifying coordinates
        """
        # Initialize the base class
        super().__init__(data, is_sparse=is_sparse)

        # Should provide either the counts, or the batch size
        assert (counts is not None) ^ (
            batch_size is not None
        ), "Provide either `counts` or `batch_size`, not both."

        # If the data is sparse, it must have a batch column and coordinates
        if is_sparse:
            has_batch_col = True
            coord_cols = COORD_COLS

        # If the number of batches is not provided, get it from the counts
        if batch_size is None:
            batch_size = len(counts)

        # If the counts are not provided, must build them once
        if counts is None:
            # Define the array functions depending on the input type
            assert has_batch_col, "Cannot get the counts without a batch column."
            ref = data if not is_sparse else data.C
            counts = self.get_counts(ref[:, BATCH_COL], batch_size)

        # Cast
        counts = self._as_long(counts)
        assert self._sum(counts) == len(
            data
        ), "The `counts` provided do not add up to the tensor length."

        # Get the boundaries between entries in the batch
        edges = self.get_edges(counts)

        # Store the attributes
        self.data = data
        self.counts = counts
        self.edges = edges
        self.batch_size = batch_size
        self.has_batch_col = has_batch_col
        self.coord_cols = coord_cols

    def __getitem__(self, batch_id):
        """Returns a subset of the tensor corresponding to one entry.

        Parameters
        ----------
        batch_id : int
            Entry index
        """
        # Make sure the batch_id is sensible
        if batch_id >= self.batch_size:
            raise IndexError(
                f"Index {batch_id} out of bound for a batch size "
                f"of ({self.batch_size})"
            )

        # Return
        lower, upper = self.edges[batch_id], self.edges[batch_id + 1]
        if not self.is_sparse:
            return self.data[lower:upper]
        else:
            return ME.SparseTensor(
                self.data.F[lower:upper], coordinates=self.data.C[lower:upper]
            )

    @property
    def tensor(self):
        """Alias for the underlying data stored.

        Returns
        -------
        Union[np.ndarray, torch.Tensor, ME.SparseTensor]
            Underlying tensor of data
        """
        return self.data

    @property
    def batch_ids(self):
        """Returns the batch ID of each of the elements in the tensor.

        Returns
        -------
        Union[np.ndarray, torch.Tensor]
            (N) Batch ID of each element in the tensor
        """
        return self._repeat(self._arange(self.batch_size), self.counts)

    def split(self):
        """Breaks up the tensor batch into its constituents.

        Returns
        -------
        List[Union[np.ndarray, torch.Tensor]]
            List of one tensor per entry in the batch
        """
        if not self.is_sparse:
            return self._split(self.data, self.splits)
        else:
            coords = self._split(self.data.C, self.splits)
            feats = self._split(self.data.F, self.splits)
            return [
                ME.SparseTensor(feats[i], coordinates=coords[i])
                for i in range(self.batch_size)
            ]

    def apply_mask(self, mask):
        """Apply a global mask to the underlying tensor, update batching.

        Parameters
        ----------
        mask : Union[np.ndarray, torch.Tensor]
            (N) Boolean mask to apply to the underlying tensor
        """
        # Update underlying tensor in place
        self.data = self.data[mask]

        # Update batching information
        batch_ids = self.batch_ids[mask]
        self.counts = self.get_counts(batch_ids, self.batch_size)
        self.edges = self.get_edges(self.counts)

    def merge(self, tensor_batch):
        """Merge this tensor batch with another.

        Parameters
        ----------
        tensor_batch : TensorBatch
            Other tensor batch object to merge with

        Returns
        -------
        TensorBatch
            Merged tensor batch
        """
        # Stack the tensors entry-wise in the batch
        entries = []
        for b in range(self.batch_size):
            entries.append(self[b])
            entries.append(tensor_batch[b])

        tensor = self._cat(entries)
        counts = self.counts + tensor_batch.counts

        return TensorBatch(tensor, counts)

    def to_numpy(self):
        """Cast underlying tensor to a `np.ndarray` and return a new instance.

        Returns
        -------
        TensorBatch
            New `TensorBatch` object with an underlying np.ndarray tensor.
        """
        # If the underlying data is of the right type, nothing to do
        if self.is_numpy:
            return self

        data = self.data
        if self.is_sparse:
            data = torch.cat(
                [self.data.C.to(dtype=self.data.F.dtype), self.data.F], dim=1
            )

        data = self._to_numpy(data)
        counts = self._to_numpy(self.counts)

        return TensorBatch(
            data, counts, has_batch_col=self.has_batch_col, coord_cols=self.coord_cols
        )

    def to_tensor(self, dtype=None, device=None):
        """Cast underlying tensor to a `torch.tensor` and return a new instance.

        Parameters
        ----------
        dtype : torch.dtype, optional
            Data type of the tensor to create
        device : torch.device, optional
            Device on which to put the tensor

        Returns
        -------
        TensorBatch
            New `TensorBatch` object with an underlying np.ndarray tensor.
        """
        # If the underlying data is of the right type, nothing to do
        if not self.is_numpy:
            return self

        data = self._to_tensor(self.data, dtype, device)
        counts = self._to_tensor(self.counts, dtype, device)

        return TensorBatch(
            data, counts, has_batch_col=self.has_batch_col, coord_cols=self.coord_cols
        )

    def to_cm(self, meta):
        """Converts the pixel coordinates of the tensor to cm.

        Parameters
        ----------
        meta : Meta
            Metadata information about the rasterized image
        """
        assert self.is_numpy, "Can only convert units of numpy arrays."
        self.data[:, COORD_COLS] = meta.to_cm(self.data[:, COORD_COLS], center=True)

    def to_px(self, meta):
        """Converts the coordinates of the tensor to pixel indexes.

        Parameters
        ----------
        meta : Meta
            Metadata information about the rasterized image
        """
        assert self.is_numpy, "Can only convert units of numpy arrays."
        self.data[:, COORD_COLS] = meta.to_px(self.data[:, COORD_COLS], floor=True)

    @classmethod
    def from_list(cls, data_list):
        """Builds a batch from a list of tensors.

        Parameters
        ----------
        data_list : List[Union[np.ndarray, torch.Tensor]]
            List of tensors, exactly one per batch
        """
        # Check that we are not fed an empty list of tensors
        assert len(
            data_list
        ), "Must provide at least one tensor to build a tensor batch"
        is_numpy = not isinstance(data_list[0], torch.Tensor)

        # Compute the counts from the input list
        counts = [len(t) for t in data_list]

        # Concatenate input
        if is_numpy:
            return cls(np.concatenate(data_list, axis=0), counts)
        else:
            return cls(torch.cat(data_list, dim=0), counts)



"""Module with a parent class of all data structures."""

from dataclasses import asdict, dataclass

import numpy as np


@dataclass(eq=False)
class DataBase:
    """Base class of all data structures.

    Defines basic methods shared by all data structures.
    """

    # Enumerated attributes
    _enum_attrs = ()

    # Fixed-length attributes as (key, size) or (key, (size, dtype)) pairs
    _fixed_length_attrs = ()

    # Variable-length attributes as (key, dtype) or (key, (width, dtype)) pairs
    _var_length_attrs = ()

    # Attributes specifying coordinates
    _pos_attrs = ()

    # Attributes specifying vector components
    _vec_attrs = ()

    # String attributes
    _str_attrs = ()

    # Boolean attributes
    _bool_attrs = ()

    # Index attributes
    _index_attrs = ()

    # Attributes to concatenate when merging objects
    _cat_attrs = ()

    # Attributes that must never be stored to file
    _skip_attrs = ()

    # Attributes that must not be stored to file when storing lite files
    _lite_skip_attrs = ()

    # Euclidean axis labels
    _axes = ("x", "y", "z")

    def __post_init__(self):
        """Immediately called after building the class attributes.

        Provides two functions:
        - Gives default values to array-like attributes. If a default value was
          provided in the attribute definition, all instances of this class
          would point to the same memory location.
        - Casts strings when they are provided as binary objects, which is the
          format one gets when loading string from HDF5 files.
        """
        # Provide default values to the variable-length array attributes
        for attr, dtype in self._var_length_attrs:
            if getattr(self, attr) is None:
                if not isinstance(dtype, tuple):
                    setattr(self, attr, np.empty(0, dtype=dtype))
                else:
                    width, dtype = dtype
                    setattr(self, attr, np.empty((0, width), dtype=dtype))

        # Provide default values to the fixed-length array attributes
        for attr, size in self._fixed_length_attrs:
            if getattr(self, attr) is None:
                if not isinstance(size, tuple):
                    dtype = np.float32
                else:
                    size, dtype = size
                setattr(self, attr, np.full(size, -np.inf, dtype=dtype))

        # Cast stored binary strings back to regular strings
        for attr in self._str_attrs:
            if isinstance(getattr(self, attr), bytes):
                setattr(self, attr, getattr(self, attr).decode())

        # Cast stored 8-bit unsigned integers back to booleans
        for attr in self._bool_attrs:
            if isinstance(getattr(self, attr), np.uint8):
                setattr(self, attr, bool(getattr(self, attr)))

    def __eq__(self, other):
        """Checks that all attributes of two class instances are the same.

        This overloads the default dataclass `__eq__` method to include an
        appopriate check for vector (numpy) attributes.

        Parameters
        ----------
        other : obj
            Other instance of the same object class

        Returns
        -------
        bool
            `True` if all attributes of both objects are identical
        """
        # Check that the two objects belong to the same class
        if self.__class__ != other.__class__:
            return False

        # Check that all base attributes are identical
        for key, value in self.__dict__.items():
            value_other = getattr(other, key)
            if np.isscalar(value):
                # For scalars, regular comparison will do
                if value_other != value:
                    return False

            else:
                # For vectors, compare all elements
                if isinstance(value, np.ndarray):
                    if value.shape != value_other.shape or (value_other != value).any():
                        return False
                else:
                    if len(value) != len(value_other) or any(
                        v1 != v2 for v1, v2 in zip(value, value_other)
                    ):
                        return False

        return True

    def set_precision(self, precision):
        """Casts all the vector attributes to a different precision.

        Parameters
        ----------
        int : default 4
            Precision in number of bytes (half=2, single=4, double=8)
        """
        assert precision in [
            2,
            4,
            8,
        ], "Set the vector attribute precision for this object."
        for attr in self.fixed_length_attrs + self.variable_length_attrs:
            val = getattr(self, attr)
            dtype = f"{val.dtype.str[:-1]}{precision}"
            setattr(self, attr, val.astype(dtype))

    def shift_indexes(self, shifts):
        """Apply offsets to index attributes in place.

        Invalid indexes (-1) are not offset to prevent making them valid.

        Parameters
        ----------
        shifts : Union[int, Dict[str, int]]
            Shift(s) to apply to the index attributes. If provided as a
            dictionary, the shifts are tied to a specific attribute.
        """
        # Dispatch
        for attr in self._index_attrs:
            value = getattr(self, attr)
            if not np.isscalar(value) or value > -1:
                if not isinstance(shifts, dict):
                    setattr(self, attr, value + shifts)
                else:
                    setattr(self, attr, value + shifts[attr])

    def as_dict(self, lite=False):
        """Returns the data class as dictionary of (key, value) pairs.

        Parameters
        ----------
        lite : bool, default False
            If `True`, the `_lite_skip_attrs` are dropped

        Returns
        -------
        dict
            Dictionary of attribute names and their values
        """
        # Build a list of attributes to skip
        if not lite:
            skip_attrs = self._skip_attrs
        else:
            skip_attrs = (*self._skip_attrs, *self._lite_skip_attrs)

        return {
            key: value for key, value in asdict(self).items() if not key in skip_attrs
        }

    def scalar_dict(self, attrs=None, lengths=None, lite=False):
        """Returns the data class attributes as a dictionary of scalars.

        This is useful when storing data classes in CSV files, which expect
        a single scalar per column in the table.

        Parameters
        ----------
        attrs : List[str], optional
            List of attribute names to include in the dictionary. If not
            specified, all the keys are included.
        lengths : Dict[str, int], optional
            Specifies the length of variable-length attributes
        lite : bool, default False
            If `True`, the `_lite_skip_attrs` are dropped
        """
        # Loop over the attributes of the data class
        lengths = lengths or {}
        scalar_dict, found = {}, []
        for attr, value in self.as_dict(lite).items():
            # If the attribute is not requested, skip
            if attrs is not None and attr not in attrs:
                continue
            else:
                found.append(attr)

            # Dispatch
            if np.isscalar(value):
                # If the attribute is a scalar, store as is
                scalar_dict[attr] = value

            elif attr in (self._pos_attrs + self._vec_attrs):
                # If the attribute is a position or vector, expand with axis
                for i, v in enumerate(value):
                    scalar_dict[f"{attr}_{self._axes[i]}"] = v

            elif attr in self.fixed_length_attrs:
                # If the attribute is a fixed-length array, expand with index
                for i, v in enumerate(value):
                    scalar_dict[f"{attr}_{i}"] = v

            elif attr in self.var_length_attrs:
                if attr in lengths:
                    # If the attribute is a variable-length array with a length
                    # provided, resize it to match that length and store it
                    for i in range(lengths[attr]):
                        if i < len(value):
                            scalar_dict[f"{attr}_{i}"] = value[i]
                        else:
                            scalar_dict[f"{attr}_{i}"] = None

                else:
                    # If the attribute is a variable-length array of
                    # indeterminate length, do not store it
                    assert attrs is None or attr not in attrs, (
                        f"Cannot cast {attr} to scalars. To cast a variable-"
                        "length array, must provide a fixed length."
                    )
                    continue

            else:
                raise ValueError(
                    f"Cannot expand the `{attr}` attribute of "
                    f"`{self.__class__.__name__}` to scalar values."
                )

        if attrs is not None and len(attrs) != len(found):
            class_name = self.__class__.__name__
            miss = list(set(attrs).difference(set(found)))
            raise AttributeError(
                f"Attribute(s) {miss} do(es) not appear in {class_name}."
            )

        return scalar_dict

    @property
    def fixed_length_attrs(self):
        """Fetches the dictionary of fixed-length array attributes as a dictionary.

        Returns
        -------
        Dict[str, int]
            Dictionary which maps fixed-length attributes onto their length
        """
        return dict(self._fixed_length_attrs)

    @property
    def var_length_attrs(self):
        """Fetches the list of variable-length array attributes as a dictionary.

        Returns
        -------
        Dict[str, type]
            Dictionary which maps variable-length attributes onto their type
        """
        return dict(self._var_length_attrs)

    @property
    def enum_attrs(self):
        """Fetches the list of enumerated attributes as a dictionary.

        Returns
        -------
        Dict[int, Dict[int, str]]
            Dictionary which maps names onto enumerator descriptors
        """
        return {key: dict(value) for key, value in self._enum_attrs}

    @property
    def index_attrs(self):
        """Fetches the list of attributes that correspond to indexes.

        Returns
        -------
        List[str]
            List of attributes that specificy indexes
        """
        return self._index_attrs

    @property
    def skip_attrs(self):
        """Fetches the list of attributes to not store to file.

        Returns
        -------
        List[str]
            List of attributes to exclude from the storage process
        """
        return self._skip_attrs

    @property
    def lite_skip_attrs(self):
        """Fetches the list of attributes to not store to lite file.

        Returns
        -------
        List[str]
            List of attributes to exclude from the storage process
        """
        return self._lite_skip_attrs


@dataclass(eq=False)
class PosDataBase(DataBase):
    """Base class of for data structures with positional attributes.

    Includes method to convert positional attributes

    Attributes
    ----------
    units : str
        Units in which the position attributes are expressed
    """

    units = "cm"

    def __post_init__(self):
        """Immediately called after building the class attributes.

        Makes sure the units are not binary and that they are recognized.
        """
        # Call the main post initialization function
        super().__post_init__()

        # Parse the units
        if isinstance(self.units, bytes):
            self.units = self.units.decode()

        assert self.units in ["cm", "px"], "Units can only be `cm` or `px`."

    def to_cm(self, meta):
        """Converts the coordinates of the positional attributes to cm.

        Parameters
        ----------
        meta : Meta
            Metadata information about the rasterized image
        """
        assert self.units != "cm", "Units already expressed in cm"
        self.units = "cm"
        for attr in self._pos_attrs:
            setattr(self, attr, meta.to_cm(getattr(self, attr)))

    def to_px(self, meta):
        """Converts the coordinates of the positional attributes to pixel.

        Parameters
        ----------
        meta : Meta
            Metadata information about the rasterized image
        """
        assert self.units != "px", "Units already expressed in pixels"
        self.units = "px"
        for attr in self._pos_attrs:
            setattr(self, attr, meta.to_px(getattr(self, attr)))



import numpy as np
import torch

from spine.data import TensorBatch
from spine.utils.globals import COORD_COLS, SHAPE_COL, VALUE_COL
from spine.utils.gnn.cluster import (
    get_cluster_dedxs_batch,
    get_cluster_directions_batch,
    get_cluster_features_batch,
    get_cluster_points_label_batch,
)
from spine.utils.gnn.network import get_cluster_edge_features_batch
from spine.utils.torch.scripts import cdist_fast

__all__ = ["ClustGeoNodeEncoder", "ClustGeoEdgeEncoder"]


class ClustGeoNodeEncoder(torch.nn.Module):
    """Produces cluster node features using hand-engineered quantities.

    The basic 16 geometric features are composed of:
    - Center (3)
    - Covariance matrix (9)
    - Principal axis (3)
    - Voxel count (1)

    The flag `add_value` adds the following 2 features:
    - Mean energy (1)
    - RMS energy (1)

    The flag `add_shape` adds the particle shape information:
    - Semantic type (1), i.e. most represented type in cluster

    The flag `add_points` adds the particle end points information
    - Start point (3)
    - End point (3)

    The flag `add_directions` adds the particle direction information
    - Start direction (3)
    - End direction (3)

    The flag `add_local_dedxs` adds the local dEdx estimate at each endpoint
    - Start dEdx (1)
    - End dEdx (1)
    """

    # Name of the node encoder (as specified in the configuration)
    name = "geometric"

    # Alternative allowed names of the node encoder
    aliases = ("geo",)

    def __init__(
        self,
        use_numpy=True,
        add_value=False,
        add_shape=False,
        add_points=False,
        add_local_dirs=False,
        dir_max_dist=5.0,
        add_local_dedxs=False,
        dedx_max_dist=5.0,
    ):
        """Initializes the geometric-based node encoder.

        Parameters
        ----------
        use_numpy : bool, default True
            Generate the features on CPU
        add_value : bool, default False
            Add mean and RMS value of pixels in the cluster
        add_shape : bool, default False
            Add the particle semantic type
        add_points : bool, default False
            Add the start/end points of the particles
        add_local_dirs : bool, default False
            Add the local direction estimates at the start and end points
        dir_max_dist : float, default 5.
            Radius around the end points included to estimate the directions
        add_local_dedxs : boo, default False
            Add the local dE/dx estimates at the start and end points
        dedx_max_dist : float, default 5.
            Readius around the end points incldued to estimate the dE/dx
        """
        # Initialize the parent class
        super().__init__()

        # Store the paramters
        self.use_numpy = use_numpy
        self.add_value = add_value
        self.add_shape = add_shape
        self.add_points = add_points
        self.add_local_dirs = add_local_dirs
        self.dir_max_dist = dir_max_dist
        self.add_local_dedxs = add_local_dedxs
        self.dedx_max_dist = dedx_max_dist

        # If the maximum distance is specified as `optimize`, optimize it
        self.opt_dir_max_dist = False
        if isinstance(self.dir_max_dist, str):
            assert self.dir_max_dist == "optimize", (
                "If specified as a string, `dir_max_dist` should "
                "only take the value 'optimize'"
            )
            self.opt_dir_max_dist = True

        # Sanity check
        assert self.add_points or (
            not self.add_local_dirs and not self.add_local_dedxs
        ), "If directions or dE/dx is requested, must also add points"

    def forward(
        self, data, clusts, coord_label=None, points=None, extra=None, **kwargs
    ):
        """Generate geometric cluster node features for one batch of data.

        Parameters
        ----------
        data : TensorBatch
            (N, 1 + D + N_f) Batch of sparse tensors
        clusts : IndexBatch
            (C) Indexes that make up each cluster
        coord_label : TensorBatch
            (P, 1 + D + 8) Label start, end, time and shape for each point
        points : TensorBatch
            (C, 6) Set of start/end points for each input cluster
        extra : TensorBatch
            (C, 1/2/3) Set of mean/rms values in the cluster and/or shape
        **kwargs : dict, optional
            Additional objects no used by this encoder

        Returns
        -------
        TensorBatch
           (C, N_c) Set of N_c features per cluster
        """
        # If features are provided directly, must ensure that the corresponding
        # flags in the configuration are as expected.
        assert (
            points is None or self.add_points
        ), "If end points are provided, `add_points` should be `True`."
        assert extra is None or (self.add_value or self.add_shape), (
            "If extra features are provided, either `add_value` or "
            "`add_shape` should be `True`."
        )
        assert not self.add_points or (
            (coord_label is not None) ^ (points is not None)
        ), ("Must provide either `coord_label` or `points` to add points, " "not both.")

        # Update the flags depending what is provided
        add_value, add_shape = self.add_value, self.add_shape
        if extra is not None:
            if self.add_value and self.add_shape:
                assert extra.shape[1] == 3
            elif self.add_value and not self.add_shape:
                assert extra.shape[1] == 2
            elif not self.add_value and self.add_shape:
                assert extra.shape[1] == 1
            add_value, add_shape = False, False

        # Extract the base geometric features
        if self.use_numpy:
            # If numpy is to be used, pass it through the Numba function
            feats = get_cluster_features_batch(
                data, clusts, add_value, add_shape
            ).tensor
        else:
            # Otherwise, use the local torch method
            feats = self.get_base_features(data, clusts, add_value, add_shape).tensor

        # Add the extra features if they were provided independantly
        if extra is not None:
            feats = torch.cat((feats, extra.tensor), dim=1)

        # Add the points
        if self.add_points:
            if points is None:
                points = get_cluster_points_label_batch(data, coord_label, clusts)

            feats = torch.cat((feats, points.tensor), dim=1)

        # Add the local directions
        if self.add_local_dirs:
            for cols in np.arange(points.tensor.shape[1]).reshape(-1, 3):
                starts = TensorBatch(points.tensor[:, cols], points.counts)
                dirs = get_cluster_directions_batch(
                    data, starts, clusts, self.dir_max_dist, self.opt_dir_max_dist
                )
                feats = torch.cat((feats, dirs.tensor), dim=1)

        # Add the local dE/dx information
        if self.add_local_dedxs:
            for cols in np.arange(points.tensor.shape[1]).reshape(-1, 3):
                starts = TensorBatch(points.tensor[:, cols], points.counts)
                dedxs = get_cluster_dedxs_batch(
                    data, starts, clusts, self.dedx_max_dist
                )
                feats = torch.cat((feats, dedxs.tensor[:, None]), dim=1)

        feats = TensorBatch(feats, clusts.counts)

        # Return
        if self.add_points:
            return feats, points

        return feats

    def get_base_features(self, data, clusts, add_value, add_shape):
        """Generate base geometric cluster node features for one batch of data.

        Parameters
        ----------
        data : TensorBatch
            (N, 1 + D + N_f) Batch of sparse tensors
        clusts : IndexBatch
            (C) Indexes that make up each cluster
        add_value : bool, default False
            Add mean and RMS value of pixels in the cluster
        add_shape : bool, default False
            Add the particle semantic type
        """
        # Get the value & semantic types
        voxels = data.tensor[:, COORD_COLS]
        values = data.tensor[:, VALUE_COL]
        sem_types = data.tensor[:, SHAPE_COL] if add_shape else None

        # Below is a torch-based implementation of cluster_features
        feats = []
        dtype, device = voxels.dtype, voxels.device
        zeros = lambda x: torch.zeros(x, dtype=dtype, device=device)
        full = lambda x, y: torch.full(x, y, dtype=dtype, device=device)
        for c in clusts.index_list:
            # Get list of voxels in the cluster
            x = voxels[c]
            size = full([1], len(c))

            # Give default values to size-1 clusters
            if len(c) < 2:
                feats_v = torch.cat((x.flatten(), zeros(12), size))
                if add_value:
                    vals = zeros(2)
                    vals[0] = values[c[0]]
                    feats_v = torch.cat((feats_v, vals))
                if add_shape and sem_types is not None:
                    shape = full([1], sem_types[c[0]])
                    feats_v = torch.cat((feats_v, shape))

                feats.append(feats_v)
                continue

            # Center data
            center = x.mean(dim=0)
            x = x - center

            # Get orientation matrix
            A = x.t().mm(x)

            # Get eigenvectors, normalize orientation matrix and
            # eigenvalues to largest. This step assumes points are not
            # superimposed, i.e. that largest eigenvalue != 0
            # w, v = torch.symeig(A, eigenvectors=True)
            w, v = torch.linalg.eigh(A, UPLO="U")
            dirwt = 1.0 - w[1] / w[2]
            B = A / w[2]

            # Get the principal direction
            v0 = v[:, 2]

            # Projection all points along the principal axis
            x0 = x.mv(v0)

            # Evaluate the distance from the points to the principal axis
            xp0 = x - torch.ger(x0, v0)
            np0 = torch.norm(xp0, dim=1)

            # Flip the principal direction if it is not pointing
            # towards the maximum spread
            sc = torch.dot(x0, np0)
            if sc < 0:
                v0 = -v0

            # Weight direction
            v0 = dirwt * v0

            # Append (center, B.flatten(), v0, size)
            feats_v = torch.cat((center, B.flatten(), v0, size))
            if add_value:
                vals = zeros(2)
                vals[0] = values[c].mean()
                vals[1] = values[c].std()
                feats_v = torch.cat((feats_v, vals))
            if add_shape:
                shape = full([1], sem_types[c].mode())
                feats_v = torch.cat((feats_v, shape))

            feats.append(feats_v)

        # Return
        if len(feats) == 0:
            return TensorBatch(torch.stack(feats, dim=0), clusts.counts)
        else:
            return TensorBatch(zeros((0, 16)), clusts.counts)


class ClustGeoEdgeEncoder(torch.nn.Module):
    """Produces cluster edge features using hand-engineered quantities.

    The basic 19 geometric features are composed of:
    - Position of the voxel in the first cluster closest to the second (3)
    - Position of the voxel in the second cluster closest to the first (3)
    - Displacement vector from the first to the second point defined above (3)
    - Length of the displacement vector (1)
    - Outer product of the displacement vector (9)
    """

    # Name of the edge encoder (as specified in the configuration)
    name = "geometric"

    # Alternative allowed names of the edge encoder
    aliases = ("geo",)

    def __init__(self, use_numpy=True):
        """Initializes the geometric-based node encoder.

        Parameters
        ----------
        use_numpy : bool, default True
            Generate the features on CPU
        """
        # Initialize the parent class
        super().__init__()

        # Store the paramters
        self.use_numpy = use_numpy

    def forward(self, data, clusts, edge_index, closest_index=None, **kwargs):
        """Generate geometric cluster edge features for one batch of data.

        Parameters
        ----------
        data : TensorBatch
            (N, 1 + D + N_f) Batch of sparse tensors
        clusts : IndexBatch
            (C) Indexes that make up each cluster
        edge_index : EdgeIndexBatch
            Incidence map between clusters
        closest_index : Union[np.ndarray, torch.Tensor], optional
            (C, C) : Combined index of the closest pair of voxels per
            pair of clusters
        **kwargs : dict, optional
            Additional objects no used by this encoder

        Returns
        -------
        TensorBatch
           (C, N_e) Set of N_e features per edge
        """
        # Extract the base geometric features
        if self.use_numpy:
            # If numpy is to be used, pass it through the Numba function
            feats = get_cluster_edge_features_batch(
                data, clusts, edge_index, closest_index=closest_index
            ).tensor
        else:
            # Otherwise, use the local torch method
            feats = self.get_base_features(data, clusts, edge_index, closest_index)

        # If the graph is undirected, infer reciprocal features
        if not edge_index.directed:
            # Create the feature tensor of reciprocal edges
            feats_flip = feats.clone()
            feats_flip[:, :3] = feats[:, 3:6]
            feats_flip[:, 3:6] = feats[:, :3]
            feats_flip[:, 6:9] = -feats[:, 6:9]

            # Create the full feature tensor
            full_feats = torch.empty(
                (2 * feats.shape[0], feats.shape[1]),
                dtype=feats.dtype,
                device=feats.device,
            )
            full_feats[::2] = feats
            full_feats[1::2] = feats_flip

            feats = full_feats

        return TensorBatch(feats, edge_index.counts)

    def get_base_features(self, data, clusts, edge_index, closest_index=None):
        """Generate base geometric cluster node features for one batch of data.

        Parameters
        ----------
        data : TensorBatch
            (N, 1 + D + N_f) Batch of sparse tensors
        clusts : IndexBatch
            (C) Indexes that make up each cluster
        edge_index : EdgeIndexBatch
            Incidence map between clusters
        closest_index : Union[np.ndarray, torch.Tensor], optional
            (C, C) : Combined index of the closest pair of voxels per
            pair of clusters
        """
        # Get the voxel set
        voxels = data.tensor[:, COORD_COLS]

        # Here is a torch-based implementation of cluster_edge_features
        feats = []
        for e in edge_index.directed_index_t:

            # Get the voxels in the clusters connected by the edge
            x1 = voxels[clusts.index_list[e[0]]]
            x2 = voxels[clusts.index_list[e[1]]]

            # Find the closest set point in each cluster
            if closest_index is None:
                d12 = cdist_fast(x1, x2)
                imin = torch.argmin(d12)
            else:
                imin = closest_index[e[0], e[1]]

            i1, i2 = imin // len(x2), imin % len(x2)
            v1 = x1[i1, :]  # closest point in c1
            v2 = x2[i2, :]  # closest point in c2

            # Displacement
            disp = v1 - v2

            # Distance
            lend = torch.norm(disp)
            if lend > 0:
                disp = disp / lend

            # Outer product
            outer = torch.ger(disp, disp).flatten()

            feats.append(torch.cat((v1, v2, disp, lend.reshape(1), outer)))

        if len(feats) == 0:
            return torch.stack(feats, dim=0)
        else:
            return torch.zeros((0, 19))



"""Numba JIT compiled implementation of clustering routines."""

import numba as nb
import numpy as np

from .distance import METRICS, get_metric_id
from .graph import connected_components, radius_graph

__all__ = ["DBSCAN", "dbscan"]

DBSCAN_DTYPE = (
    ("eps", nb.float32),
    ("min_samples", nb.int64),
    ("metric_id", nb.int64),
    ("p", nb.float32),
)


@nb.experimental.jitclass(DBSCAN_DTYPE)
class DBSCAN:
    """Class-version of the Numba-accelerate :func:`dbscan` function.

    Attributes
    ----------
    eps : float
        Distance scale (determines neighborhood)
    min_samples : int
        Minimum number of neighbors (including oneself) to be considered
        a core point
    metric : str
        Distance metric to be used to establish neighborhood
    """

    def __init__(
        self,
        eps: nb.float32,
        min_samples: nb.int64 = 1,
        metric: nb.types.string = "euclidean",
        p: nb.int64 = 2.0,
    ):
        """Initialize the DBSCAN parameters.

        Parameters
        ----------
        eps : float
            Distance scale (determines neighborhood)
        min_samples : int
            Minimum number of neighbors (including oneself) to be considered
            a core point
        metric : str
            Distance metric to be used to establish neighborhood
        p : float, default 2.
            p-norm factor for the Minkowski metric, if used
        """
        # For Euclidean, save time by using squared Euclidean
        if metric == "euclidean":
            metric = "sqeuclidean"
            eps = eps * eps

        # Store parameters
        self.eps = eps
        self.min_samples = min_samples
        self.metric_id = get_metric_id(metric, p)
        self.p = p

    def fit_predict(self, x):
        """Runs DBSCAN on 3D points and returns the group assignments.

        Parameters
        ----------
        x : np.ndarray
            (N, 3) array of point coordinates
        eps : float
            Distance below which two points are considered neighbors
        min_samples : int
            Minimum number of neighbors for a point to be a core point
        metric : str, default 'euclidean'
            Distance metric used to compute pdist

        Returns
        -------
        np.ndarray
            (N) Group assignments
        """
        # Produce a radius graph
        edge_index = radius_graph(x, self.eps, self.metric_id, self.p)

        # Build groups
        return connected_components(
            edge_index, len(x), self.min_samples, directed=False
        )


@nb.njit(cache=True)
def dbscan(
    x: nb.float32[:, :],
    eps: nb.float32,
    min_samples: nb.int64 = 1,
    metric_id: nb.int64 = METRICS["euclidean"],
    p: nb.float32 = 2.0,
) -> nb.float32[:]:
    """Runs DBSCAN on 3D points and returns the group assignments.

    Parameters
    ----------
    x : np.ndarray
        (N, 3) array of point coordinates
    eps : float
        Distance below which two points are considered neighbors
    min_samples : int
        Minimum number of neighbors for a point to be a core point
    metric : str, default 'euclidean'
        Distance metric used to compute pdist
    p : float, default 2.
        p-norm factor for the Minkowski metric, if used

    Returns
    -------
    np.ndarray
        (N) Group assignments
    """
    # Produce a radius graph
    edge_index = radius_graph(x, eps, metric_id, p)

    # Build groups
    return connected_components(edge_index, len(x), min_samples, directed=False)



"""Manages the operation of analysis scripts."""

from collections import OrderedDict, defaultdict
from copy import deepcopy

import numpy as np

from spine.utils.stopwatch import StopwatchManager

from .factories import ana_script_factory


class AnaManager:
    """Manager class to initialize and execute analysis scripts.

    Analysis scripts use the output of the reconstruction chain and the
    post-processors and produce simple CSV files.

    It loads all the analysis scripts and feeds them data. It initializes
    CSV writers needed to store the output of the analysis scripts.
    """

    def __init__(self, cfg, log_dir=None, prefix=None):
        """Initialize the analysis manager.

        Parameters
        ----------
        cfg : dict
            Analysis script configurations
        log_dir : str
            Output CSV file directory (shared with driver log)
        prefix : str, optional
            Input file prefix. If requested, it will be used to prefix
            all the output CSV files.
        """
        # Parse the analysis block configuration
        self.parse_config(log_dir, prefix, **cfg)

    def parse_config(
        self, log_dir, prefix, overwrite=None, prefix_output=False, **modules
    ):
        """Parse the analysis tool configuration.

        Parameters
        ----------
        log_dir : str
            Output CSV file directory (shared with driver log)
        prefix : str
            Input file prefix. If requested, it will be used to prefix
            all the output CSV files.
        overwrite : bool, optional
            If `True`, overwrite the CSV logs if they already exist
        prefix_output : bool, optional
            If `True`, will prefix the output CSV names with the input file name
        **modules : dict
            List of analysis script modules
        """
        # Loop over the analyzer modules and get their priorities
        modules = deepcopy(modules)
        keys = np.array(list(modules.keys()))
        priorities = -np.ones(len(keys), dtype=np.int32)
        for i, k in enumerate(keys):
            if "priority" in modules[k]:
                priorities[i] = modules[k].pop("priority")

        # Only use the prefix if the output is to be prefixed
        if not prefix_output:
            prefix = None

        # Add the modules to a processor list in decreasing order of priority
        self.watch = StopwatchManager()
        self.modules = OrderedDict()
        keys = keys[np.argsort(-priorities)]
        for k in keys:
            # Profile the module
            self.watch.initialize(k)

            # Append
            self.modules[k] = ana_script_factory(
                k, modules[k], overwrite, log_dir, prefix
            )

    def __call__(self, data):
        """Pass one batch of data through the analysis scripts

        Parameters
        ----------
        data : dict
            Dictionary of data products
        """
        # Loop over the analysis script modules
        single_entry = np.isscalar(data["index"])
        for key, module in self.modules.items():
            # Run the analysis script on each entry
            self.watch.start(key)
            if single_entry:
                num_entries = 1
                result = module(data)

            else:
                num_entries = len(data["index"])
                result = defaultdict(list)
                for entry in range(num_entries):
                    result_e = module(data, entry)
                    if result_e is not None:
                        for k, v in result_e.items():
                            result[k].append(v)

            self.watch.stop(key)

            # Update the input dictionary
            if result is not None:
                for key, val in result.items():
                    if not single_entry:
                        assert len(val) == num_entries, (
                            f"The number {key} ({len(val)}) does not match "
                            f"the number of entries ({num_entries})."
                        )
                    data[key] = val



"""Visualization tools for SPINE data and results.

This module provides comprehensive visualization capabilities for neutrino physics data,
ML model outputs, and analysis results using Plotly for interactive 3D visualization.

**Core Visualization Functions:**
- `point`: General voxel and point cloud visualization (`scatter_points`)
- `particle`: Particle trajectory and property visualization
- `cluster`: Cluster analysis and display (`scatter_clusters`)
- `box`: Bounding box and region-of-interest visualization (`scatter_boxes`)
- `arrow`: Directional vector and momentum visualization

**Network & Graph Visualization:**
- `network`: Graph neural network topology visualization
  - `network_topology`: Display graph structure and connections
  - `network_schematic`: Simplified network diagrams

**Analysis & Output Visualization:**
- `out`: Model output and prediction visualization
- `evaluation`: Performance metrics and evaluation plots
- `train`: Training progress and loss visualization

**Geometry & Detector:**
- `geo`: Detector geometry and coordinate system visualization

**Layout & Styling:**
- `layout`: Pre-configured Plotly layouts (`plotly_layout3d`) for consistent styling

**Key Features:**
- Interactive 3D visualization with Plotly
- Automatic color coding and legends
- Support for large datasets with efficient rendering
- Detector geometry integration
- Truth vs. prediction comparison plots
- Export capabilities for publication-quality figures

**Example Usage:**
```python
from spine.vis import scatter_points, scatter_clusters, layout3d

# Visualize detector hits
fig = scatter_points(coordinates, features, color='energy')
fig.update_layout(layout3d)

# Visualize clustering results
fig = scatter_clusters(data, cluster_labels, method='dbscan')
```
"""

from .arrow import *
from .box import *
from .cluster import *
from .evaluation import *
from .geo import *
from .layout import *
from .network import *
from .out import *
from .particle import *
from .point import *
from .train import *



"Local implementations of the AdaBound optimizers."

import math

from ..conditional import TORCH_AVAILABLE, torch

if not TORCH_AVAILABLE:
    raise ImportError(
        "PyTorch is required for AdaBound optimizers. "
        "Install with: pip install spine-ml[model]"
    )

from torch.optim import Optimizer

__all__ = ["AdaBound", "AdaBoundW"]


class AdaBound(Optimizer):
    """Implements AdaBound algorithm.

    It has been proposed in `Adaptive Gradient Methods with Dynamic Bound of
    Learning Rate`.

    .. Adaptive Gradient Methods with Dynamic Bound of Learning Rate:
        https://openreview.net/forum?id=Bkg3g2R9FX
    """

    name = "ababound"

    def __init__(
        self,
        params,
        lr=1e-3,
        betas=(0.9, 0.999),
        final_lr=0.1,
        gamma=1e-3,
        eps=1e-8,
        weight_decay=0,
        amsbound=False,
    ):
        """Initialize the AdaBound optimizer

        Parameters
        ----------
        params : iterable
            Iiterable of parameters to optimize or dicts defining
            parameter groups
        lr : float, default 1e-3
            Adam learning rate
        betas : Tuple[float, float], default (0.9, 0.999)
            Coefficients used for computing running averages of gradient and
            its square
        final_lr : float, default 0.1
            Final (SGD) learning rate
        gamma : float, default 1e-3
            Convergence speed of the bound functions
        eps : float, default 1e-8
            Term added to the denominator to improve numerical stability
        weight_decay : float, default 0.
            Weight decay (L2 penalty)
        amsbound : bool, default False
            If `True`, use the AMSBound variant of this algorithm
        """
        # Check on the input parameters, intialize the parent class
        if not 0.0 <= lr:
            raise ValueError(f"Invalid learning rate: {lr}")
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
        if not 0.0 <= final_lr:
            raise ValueError("Invalid final learning rate: {}".format(final_lr))
        if not 0.0 <= gamma < 1.0:
            raise ValueError("Invalid gamma parameter: {}".format(gamma))
        defaults = dict(
            lr=lr,
            betas=betas,
            final_lr=final_lr,
            gamma=gamma,
            eps=eps,
            weight_decay=weight_decay,
            amsbound=amsbound,
        )
        super().__init__(params, defaults)

        # Store base learning rates
        self.base_lrs = list(map(lambda group: group["lr"], self.param_groups))

    def __setstate__(self, state):
        """Load optimizer state.

        Parameters
        ----------
        state : object
            Optimizer state dictionary
        """
        # Simply call the parent function
        super().__setstate__(state)
        for group in self.param_groups:
            group.setdefault("amsbound", False)

    def step(self, closure=None):
        """Performs a single optimization step.

        Parameters
        ----------
        closure : callable, optional
            A closure that reevaluates the model and returns the loss
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group, base_lr in zip(self.param_groups, self.base_lrs):
            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError(
                        "Adam does not support sparse gradients, please consider SparseAdam instead"
                    )
                amsbound = group["amsbound"]

                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state["step"] = 0
                    # Exponential moving average of gradient values
                    state["exp_avg"] = torch.zeros_like(p.data)
                    # Exponential moving average of squared gradient values
                    state["exp_avg_sq"] = torch.zeros_like(p.data)
                    if amsbound:
                        # Maintains max of all exp. moving avg. of sq. grad. values
                        state["max_exp_avg_sq"] = torch.zeros_like(p.data)

                exp_avg, exp_avg_sq = state["exp_avg"], state["exp_avg_sq"]
                if amsbound:
                    max_exp_avg_sq = state["max_exp_avg_sq"]
                beta1, beta2 = group["betas"]

                state["step"] += 1

                if group["weight_decay"] != 0:
                    grad = grad.add(group["weight_decay"], p.data)

                # Decay the first and second moment running average coefficient
                exp_avg.mul_(beta1).add_(1 - beta1, grad)
                exp_avg_sq.mul_(beta2).addcmul_(1 - beta2, grad, grad)
                if amsbound:
                    # Maintains the maximum of all 2nd moment running avg. till now
                    torch.max(max_exp_avg_sq, exp_avg_sq, out=max_exp_avg_sq)
                    # Use the max. for normalizing running avg. of gradient
                    denom = max_exp_avg_sq.sqrt().add_(group["eps"])
                else:
                    denom = exp_avg_sq.sqrt().add_(group["eps"])

                bias_correction1 = 1 - beta1 ** state["step"]
                bias_correction2 = 1 - beta2 ** state["step"]
                step_size = group["lr"] * math.sqrt(bias_correction2) / bias_correction1

                # Applies bounds on actual learning rate
                # lr_scheduler cannot affect final_lr, this is a workaround to apply lr decay
                final_lr = group["final_lr"] * group["lr"] / base_lr
                lower_bound = final_lr * (1 - 1 / (group["gamma"] * state["step"] + 1))
                upper_bound = final_lr * (1 + 1 / (group["gamma"] * state["step"]))
                step_size = torch.full_like(denom, step_size)
                step_size.div_(denom).clamp_(lower_bound, upper_bound).mul_(exp_avg)

                p.data.add_(-step_size)

        return loss


class AdaBoundW(Optimizer):
    """Implements AdaBound algorithm with Decoupled Weight Decay (arxiv.org/abs/1711.05101)
    It has been proposed in `Adaptive Gradient Methods with Dynamic Bound of Learning Rate`_.
    Arguments:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float, optional): Adam learning rate (default: 1e-3)
        betas (Tuple[float, float], optional): coefficients used for computing
            running averages of gradient and its square (default: (0.9, 0.999))
        final_lr (float, optional): final (SGD) learning rate (default: 0.1)
        gamma (float, optional): convergence speed of the bound functions (default: 1e-3)
        eps (float, optional): term added to the denominator to improve
            numerical stability (default: 1e-8)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        amsbound (boolean, optional): whether to use the AMSBound variant of this algorithm
    .. Adaptive Gradient Methods with Dynamic Bound of Learning Rate:
        https://openreview.net/forum?id=Bkg3g2R9FX
    """

    def __init__(
        self,
        params,
        lr=1e-3,
        betas=(0.9, 0.999),
        final_lr=0.1,
        gamma=1e-3,
        eps=1e-8,
        weight_decay=0,
        amsbound=False,
    ):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
        if not 0.0 <= final_lr:
            raise ValueError("Invalid final learning rate: {}".format(final_lr))
        if not 0.0 <= gamma < 1.0:
            raise ValueError("Invalid gamma parameter: {}".format(gamma))
        defaults = dict(
            lr=lr,
            betas=betas,
            final_lr=final_lr,
            gamma=gamma,
            eps=eps,
            weight_decay=weight_decay,
            amsbound=amsbound,
        )
        super(AdaBoundW, self).__init__(params, defaults)

        self.base_lrs = list(map(lambda group: group["lr"], self.param_groups))

    def __setstate__(self, state):
        super(AdaBoundW, self).__setstate__(state)
        for group in self.param_groups:
            group.setdefault("amsbound", False)

    def step(self, closure=None):
        """Performs a single optimization step.
        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group, base_lr in zip(self.param_groups, self.base_lrs):
            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError(
                        "Adam does not support sparse gradients, please consider SparseAdam instead"
                    )
                amsbound = group["amsbound"]

                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state["step"] = 0
                    # Exponential moving average of gradient values
                    state["exp_avg"] = torch.zeros_like(p.data)
                    # Exponential moving average of squared gradient values
                    state["exp_avg_sq"] = torch.zeros_like(p.data)
                    if amsbound:
                        # Maintains max of all exp. moving avg. of sq. grad. values
                        state["max_exp_avg_sq"] = torch.zeros_like(p.data)

                exp_avg, exp_avg_sq = state["exp_avg"], state["exp_avg_sq"]
                if amsbound:
                    max_exp_avg_sq = state["max_exp_avg_sq"]
                beta1, beta2 = group["betas"]

                state["step"] += 1

                # Decay the first and second moment running average coefficient
                exp_avg.mul_(beta1).add_(1 - beta1, grad)
                exp_avg_sq.mul_(beta2).addcmul_(1 - beta2, grad, grad)
                if amsbound:
                    # Maintains the maximum of all 2nd moment running avg. till now
                    torch.max(max_exp_avg_sq, exp_avg_sq, out=max_exp_avg_sq)
                    # Use the max. for normalizing running avg. of gradient
                    denom = max_exp_avg_sq.sqrt().add_(group["eps"])
                else:
                    denom = exp_avg_sq.sqrt().add_(group["eps"])

                bias_correction1 = 1 - beta1 ** state["step"]
                bias_correction2 = 1 - beta2 ** state["step"]
                step_size = group["lr"] * math.sqrt(bias_correction2) / bias_correction1

                # Applies bounds on actual learning rate
                # lr_scheduler cannot affect final_lr, this is a workaround to apply lr decay
                final_lr = group["final_lr"] * group["lr"] / base_lr
                lower_bound = final_lr * (1 - 1 / (group["gamma"] * state["step"] + 1))
                upper_bound = final_lr * (1 + 1 / (group["gamma"] * state["step"]))
                step_size = torch.full_like(denom, step_size)
                step_size.div_(denom).clamp_(lower_bound, upper_bound).mul_(exp_avg)

                if group["weight_decay"] != 0:
                    decayed_weights = torch.mul(p.data, group["weight_decay"])
                    p.data.add_(-step_size)
                    p.data.sub_(decayed_weights)
                else:
                    p.data.add_(-step_size)

        return loss



"""Module that contains parsers that do not fit in other categories.

Contains the following parsers:
- :class:`Meta2DParser`
- :class:`Meta3DParser`
- :class:`RunInfoParser`
- :class:`CRTHitParser`
- :class:`TriggerParser`
"""

import numpy as np

from spine.data import CRTHit, Flash, Meta, RunInfo, Trigger
from spine.geo import GeoManager
from spine.utils.conditional import larcv
from spine.utils.optical import FlashMerger

from .base import ParserBase
from .data import ParserObjectList

__all__ = [
    "MetaParser",
    "RunInfoParser",
    "FlashParser",
    "CRTHitParser",
    "TriggerParser",
]


class MetaParser(ParserBase):
    """Get the metadata information to translate into real world coordinates.

    Each entry in a dataset is a cube, where pixel/voxel coordinates typically
    go from 0 to some integer N in each dimension. If you wish to translate
    these pixel/voxel coordinates back into real world coordinates, you can use
    the output of this parser to compute it.

    .. code-block. yaml

        schema:
          meta:
            parser: meta
            sparse_event: sparse3d_pcluster
    """

    # Name of the parser (as specified in the configuration)
    name = "meta"

    # Alternative allowed names of the parser
    aliases = ("meta2d", "meta3d")

    # Type of object(s) returned by the parser
    returns = "object"

    # Overlay strategy for the objects returned by the parser
    overlay = "match"

    def __call__(self, trees):
        """Parse one entry.

        Parameters
        ----------
        trees : dict
            Dictionary which maps each data product name to a LArCV object
        """
        return self.process(**self.get_input_data(trees))

    def __init__(self, projection_id=None, **kwargs):
        """Initialize the parser.

        Parameters
        ----------
        projection_id : int, optional
            Projection ID to get the 2D image from (if fetching from 2D)
        **kwargs : dict, optional
            data product arguments to be passed to the `process` function
        """
        # Initialize the parent class
        super().__init__(**kwargs)

        # Store the revelant attributes
        self.projection_id = projection_id

    def process(self, sparse_event=None, cluster_event=None):
        """Fetches the metadata from one object that has it.

        Parameters
        ----------
        sparse_event : Union[larcv.EventSparseTensor2D
                             larcv.EventSparseTensor3D], optional
            Tensor which contains the metadata information as an attribute
        cluster_event : Union[larcv.EventClusterPixel2D,
                              larcv.EventClusterVoxel3D], optional
            Cluster which contains the metadata information as an attribute

        Returns
        -------
        Meta
            Metadata information for one image
        """
        # Check on the input, pick a source for the metadata
        if (sparse_event is None) == (cluster_event is None):
            raise ValueError(
                "Must specify exactly one of `sparse_event` or `cluster_event`."
            )

        ref_event = sparse_event if sparse_event is not None else cluster_event
        assert ref_event is not None  # For the type checker's benefitV

        # Fetch a specific projection, if needed
        if isinstance(
            ref_event, (larcv.EventSparseTensor2D, larcv.EventClusterPixel2D)
        ):
            ref_event = ref_event.sparse_tensor_2d(self.projection_id)

        return Meta.from_larcv(ref_event.meta())


class RunInfoParser(ParserBase):
    """Parse run information (run, subrun, event number).

    .. code-block. yaml

        schema:
          run_info:
            parser: run_info
            sparse_event: sparse3d_pcluster
    """

    # Name of the parser (as specified in the configuration)
    name = "run_info"

    # Type of object(s) returned by the parser
    returns = "object"

    # Overlay strategy for the objects returned by the parser
    overlay = "cat"

    def __call__(self, trees):
        """Parse one entry.

        Parameters
        ----------
        trees : dict
            Dictionary which maps each data product name to a LArCV object
        """
        return self.process(**self.get_input_data(trees))

    def process(self, sparse_event=None, cluster_event=None):
        """Fetches the run information from one object that has it.

        Parameters
        ----------
        sparse_event : Union[larcv.EventSparseTensor2D
                             larcv.EventSparseTensor3D], optional
            Tensor which contains the run information as an attribute
        cluster_event : Union[larcv.EventClusterPixel2D,
                              larcv.EventClusterVoxel3D], optional
            Cluster which contains the run information as an attribute

        Returns
        -------
        RunInfo
            Run information object
        """
        # Check on the input, pick a source for the run information
        assert (sparse_event is not None) ^ (
            cluster_event is not None
        ), "Must specify either `sparse_event` or `cluster_event`."
        ref_event = sparse_event if sparse_event is not None else cluster_event

        return RunInfo.from_larcv(ref_event)


class FlashParser(ParserBase):
    """Copy construct Flash and return an array of `Flash`.

    This parser also takes care of flashes that have been split between their
    respective optical volumes, provided a `flash_event_list`. This parser
    assumes that the trees are provided in order of the volume ID they
    correspond to.

    .. code-block. yaml
        schema:
          flashes:
            parser: flash
            flash_event_list:
              - flash_cryoE
              - flash_cryoW
    """

    # Name of the parser (as specified in the configuration)
    name = "flash"

    # Alternative allowed names of the parser
    aliases = ("opflash",)

    # Type of object(s) returned by the parser
    returns = "object_list"

    def __init__(self, merge=None, **kwargs):
        """Initialize the flash parser.

        Parameters
        ----------
        merge : dict, optional
            Flash merging configuration
        **kwargs : dict, optional
            data product arguments to be passed to the `process` function
        """
        # Initialize the parent class
        super().__init__(**kwargs)

        # Initialize the flash merging class, if needed
        self.merger = None
        if merge is not None:
            self.merger = FlashMerger(**merge)

        # Fetch the geometry information to resize the flash objects
        geo = GeoManager.get_instance()
        if geo.optical is not None:
            self.num_channels_per_volume = geo.optical.num_channels_per_volume
        else:
            raise ValueError(
                "Optical geometry not found, required to resize the flash objects."
            )

    def __call__(self, trees):
        """Parse one entry.

        Parameters
        ----------
        trees : dict
            Dictionary which maps each data product name to a LArCV object
        """
        return self.process(**self.get_input_data(trees))

    def process(self, flash_event=None, flash_event_list=None):
        """Fetches the list of optical flashes.

        Parameters
        -------------
        flash_event : larcv.EventFlash, optional
            Optical flash event which contains a list of flash objects
        flash_event_list : List[larcv.EventFlash], optional
            List of optical flash events, each a list of flash objects

        Returns
        -------
        List[Flash]
            List of optical flash objects
        """
        # Check on the input
        assert (flash_event is not None) ^ (
            flash_event_list is not None
        ), "Must specify either `flash_event` or `flash_event_list`."

        # Parse flash objects
        flashes = []
        if flash_event is not None:
            # If there is a single flash event, parse it as is
            flash_list = flash_event.as_vector()
            flashes = [Flash.from_larcv(larcv.Flash(f)) for f in flash_list]

        elif flash_event_list is not None:
            # Otherwise, set the volume ID of the flash to the source index
            # and count the flash index from 0 to the largest number
            idx = 0
            for volume_id, flash_event in enumerate(flash_event_list):
                for f in flash_event.as_vector():
                    # Cast and update attributes
                    flash = Flash.from_larcv(f)
                    flash.id = idx
                    flash.volume_id = volume_id

                    # Append, increment counter
                    flashes.append(flash)
                    idx += 1

        # Resize the flash objects to match the geometry
        for flash in flashes:
            pe_per_ch = np.zeros(
                self.num_channels_per_volume, dtype=flash.pe_per_ch.dtype
            )
            if len(flash.pe_per_ch) > len(pe_per_ch):
                # If the flash spans > 1 optical volume, reshape
                lower = flash.volume_id * len(pe_per_ch)
                upper = (flash.volume_id + 1) * len(pe_per_ch)
                pe_per_ch = flash.pe_per_ch[lower:upper]

            elif len(flash.pe_per_ch) < len(pe_per_ch):
                # Pad if it does not fill the full length
                pe_per_ch[: len(flash.pe_per_ch)] = flash.pe_per_ch

            else:
                pe_per_ch = flash.pe_per_ch

            flash.pe_per_ch = pe_per_ch

        # If requested, merge flashes which match in time
        if self.merger is not None:
            flashes, _ = self.merger(flashes)

        return ParserObjectList(flashes, Flash())


class CRTHitParser(ParserBase):
    """Copy construct CRTHit and return an array of `CRTHit`.

    .. code-block. yaml
        schema:
          crthits:
            parser: crthit
            crthit_event: crthit_crthit
    """

    # Name of the parser (as specified in the configuration)
    name = "crthit"

    # Type of object(s) returned by the parser
    returns = "object_list"

    def __call__(self, trees):
        """Parse one entry.

        Parameters
        ----------
        trees : dict
            Dictionary which maps each data product name to a LArCV object
        """
        return self.process(**self.get_input_data(trees))

    def process(self, crthit_event):
        """Fetches the list of CRT hits.

        Parameters
        ----------
        crthit_event : larcv.CRTHitEvent

        Returns
        -------
        List[CRTHit]
            List of CRT hit objects
        """
        # Output as a list of LArCV CRT hit objects
        crthit_list = crthit_event.as_vector()
        crthits = [CRTHit.from_larcv(larcv.CRTHit(c)) for c in crthit_list]

        return ParserObjectList(crthits, CRTHit())


class TriggerParser(ParserBase):
    """Copy construct Trigger and return a `Trigger`.

    .. code-block. yaml
        schema:
          trigger:
            parser: trigger
            trigger_event: trigger_base
    """

    # Name of the parser (as specified in the configuration)
    name = "trigger"

    # Type of object(s) returned by the parser
    returns = "object"

    # Overlay strategy for the objects returned by the parser
    overlay = "cat"

    def __call__(self, trees):
        """Parse one entry.

        Parameters
        ----------
        trees : dict
            Dictionary which maps each data product name to a LArCV object
        """
        return self.process(**self.get_input_data(trees))

    def process(self, trigger_event):
        """Fetches the trigger information.

        Parameters
        ----------
        trigger_event : larcv.TriggerEvent

        Returns
        -------
        Trigger
            Trigger object
        """
        # Output as a trigger objects
        trigger = Trigger.from_larcv(larcv.Trigger(trigger_event))

        return trigger



"""Construct a geometry module class from its name."""

from pathlib import Path
from typing import Dict, Optional, Union

import yaml

from .base import Geometry

# Get config directory relative to this module
GEO_CONFIG_DIR = Path(__file__).parent / "config"

__all__ = ["geo_factory"]


def geo_dict() -> Dict[Path, Dict[str, str]]:
    """Builds a dictionary of available geometry modules.

    Returns
    -------
    dict
        Dictionary of available geometry modules
    """
    # Gather all geometry yaml files from the config directory
    options = {}
    for path in GEO_CONFIG_DIR.glob("*/*_geometry.yaml"):
        with open(path, "r", encoding="utf-8") as f:
            cfg = yaml.safe_load(f)

        options[path] = {k: cfg[k] for k in ("name", "tag", "version")}
        options[path]["version"] = str(float(options[path]["version"]))

    return options


def geo_factory(
    detector: str,
    tag: Optional[str] = None,
    version: Optional[Union[str, int, float]] = None,
) -> Geometry:
    """Instantiates a geometry module from a name or

    Parameters
    ----------
    detector : str
        Name of the detector (e.g. "icarus", "2x2", "ndlar", "protodune-vd", etc.)
    tag : str, optional
        Geometry tag (e.g. "sbndv2", "mr5", etc.)
    version : str, optional
        Geometry version (e.g. "1", "6.5", etc.)

    Returns
    -------
    object
         Initialized geometry object
    """
    # Find a geometry configuration that matches the requested parameters
    options = geo_dict()
    paths, tags, versions = [], [], []
    for path, cfg in options.items():
        # If the detector name matches, store the path
        if cfg["name"].lower() == detector.lower():
            paths.append(path)
            tags.append(cfg.get("tag", None))
            versions.append(cfg.get("version", None))

    if len(paths) == 0:
        raise ValueError(f"No geometry found for detector '{detector}'.")

    # If a tag is specified, must find the exact tag or throw
    file_path = ""
    if tag is not None:
        if tag in tags:
            index = tags.index(tag)
            assert version is None or version == versions[index], (
                f"Geometry version '{version}' does not match found version "
                f"'{versions[index]}' for detector '{detector}' with tag '{tag}'."
            )
            file_path = paths[index]
        else:
            raise ValueError(
                f"No geometry found for detector '{detector}' with tag '{tag}'. "
                f"Available tags are: {set(tags)}"
            )

    # If a version is specified, must match major revision if it is the only
    # one specified or both if major and minor are specified
    elif version is not None:
        version_parts = str(version).split(".")
        matched = False
        for i, ver in enumerate(versions):
            ver_parts = ver.split(".")
            if len(version_parts) == 1:
                # Only major version specified
                if version_parts[0] == ver_parts[0]:
                    file_path = paths[i]
                    matched = True
                    break
            elif len(version_parts) == 2:
                # Major and minor version specified
                if (version_parts[0] == ver_parts[0]) and (
                    version_parts[1] == ver_parts[1]
                ):
                    file_path = paths[i]
                    matched = True
                    break
        if not matched:
            raise ValueError(
                f"No geometry found for detector '{detector}' with version '{version}'. "
                f"Available versions are: {set(versions)}"
            )

    # If no tag or version is specified, return the most recent version
    else:
        index = versions.index(max(versions))
        file_path = paths[index]

    # Parse configuration file as a dictionary
    with open(file_path, "r", encoding="utf-8") as f:
        cfg = yaml.safe_load(f)

    # Instantiate the geometry module
    return Geometry(**cfg)



"""Contains a reader class dedicated to loading data from LArCV files."""

from typing import Any, Dict, List, Optional, Union

import numpy as np

from spine.utils.conditional import ROOT
from spine.utils.docstring import inherit_docstring
from spine.utils.logger import logger

from .base import ReaderBase

__all__ = ["LArCVReader"]


@inherit_docstring(ReaderBase)
class LArCVReader(ReaderBase):
    """Class which reads information stored in LArCV files.

    This class inherits from the :class:`ReaderBase` class. It provides
    methods to load LArCV2 files and extract their data products:
      - EventSparseTensor: voxel IDs and their values
      - EventClusterSparseTensor: list of sparse tensors
      - EventParticle: list of Geant4 particle information
      - EventNeutrino: list of generstor neutrino information
      - EventFlash: list of optical flashes information
      - EventCRTHit: list of cosmic-ray tagger hits
      - EventTrigger: trigger information

    It builds a TChain from the list of files provided with the appropriate
    trees corresponding to each of the requested data products.
    """

    name = "larcv"

    def __init__(
        self,
        file_keys: Optional[Union[str, List[str]]] = None,
        file_list: Optional[str] = None,
        tree_keys: Optional[List[str]] = None,
        limit_num_files: Optional[int] = None,
        max_print_files: int = 10,
        n_entry: Optional[int] = None,
        n_skip: Optional[int] = None,
        entry_list: Optional[List[int]] = None,
        skip_entry_list: Optional[List[int]] = None,
        run_event_list: Optional[List[List[int]]] = None,
        skip_run_event_list: Optional[List[List[int]]] = None,
        create_run_map: bool = False,
        run_info_key: Optional[str] = None,
        allow_missing: bool = False,
    ):
        """Initialize the LArCV file reader.

        Parameters
        ----------
        file_keys : Union[str, List[str]], optional
            Path or list of paths to the LArCV files to be read
        file_list : str, optional
            Path to a text file containing a list of file paths to be read
        tree_keys : List[str]
            List of data keys to load from the LArCV files
        limit_num_files : Optional[int], optional
            Integer limiting number of files to be taken per data directory
        max_print_files : int, default 10
            Maximum number of loaded file names to be printed
        n_entry : Optional[int], optional
            Maximum number of entries to load
        n_skip : Optional[int], optional
            Number of entries to skip at the beginning
        entry_list : Optional[List[int]]
            List of integer entry IDs to add to the index
        skip_entry_list : Optional[List[int]]
            List of integer entry IDs to skip from the index
        run_event_list: Optional[List[List[int]]], optional
            List of (run, subrun, event) triplets to add to the index
        skip_run_event_list: Optional[List[List[int]]], optional
            List of (run, subrun, event) triplets to skip from the index
        create_run_map : bool, default False
            Initialize a map between (run, subrun, event) triplets and entries.
            For large files, this can be quite expensive (must load every entry).
        run_info_key : Optional[str], optional
            Key of the tree in the file to get the run information from
        allow_missing : bool, default False
            If `True`, allows missing entries in the entry or event list
        """
        # Process the file_paths
        self.process_file_paths(file_keys, file_list, limit_num_files, max_print_files)
        assert (
            tree_keys is not None and len(tree_keys) > 0
        ), "No input `tree_keys` provided, abort."

        # If an entry list is requested based on run/subrun/event ID, create map
        if run_event_list is not None or skip_run_event_list is not None:
            create_run_map = True

        # Prepare TTrees and load files
        self.num_entries = -1
        self.trees = {}
        self.trees_ready = False
        self.file_offsets = np.empty(len(self.file_paths), dtype=np.int64)
        file_counts = []
        for key in tree_keys:
            # Check data TTree exists, and entries are identical across all
            # trees. Do not register these TTrees in yet in order to support
            # > 1 workers by the DataLoader object downstrean.
            logger.info("Loading tree %s", key)
            chain = ROOT.TChain(f"{key}_tree")  # pylint: disable=E1101
            for i, f in enumerate(self.file_paths):
                self.file_offsets[i] = chain.GetEntries()
                chain.AddFile(f)
                if key == tree_keys[0]:
                    count = chain.GetEntries() - self.file_offsets[i]
                    file_counts.append(count)

            if self.num_entries >= 0:
                assert self.num_entries == chain.GetEntries(), (
                    f"Mismatch between the number of entries for {key} "
                    f"({chain.GetEntries()}) and the number of entries "
                    f"in other data products ({self.num_entries})."
                )
            else:
                self.num_entries = chain.GetEntries()

            self.trees[key] = None
        logger.info("")

        # Dump the number of entries to load
        logger.info("Total number of entries in the file(s): %d\n", self.num_entries)

        # Build a file index
        self.file_index = np.repeat(np.arange(len(self.file_paths)), file_counts)

        # If requested, must extract the run information for each entry
        if create_run_map:
            # Initialize the TChain object
            assert run_info_key is not None and run_info_key in tree_keys, (
                "Must provide the `run_info_key` if a run map is needed. "
                "The key must appear in the list of `tree_keys`"
            )
            chain = ROOT.TChain(f"{run_info_key}_tree")  # pylint: disable=E1101
            for f in self.file_paths:
                chain.AddFile(f)

            # Loop over entries
            self.run_info = []
            for i in range(self.num_entries):
                chain.GetEntry(i)
                info = getattr(chain, f"{run_info_key}_branch")
                self.run_info.append((info.run(), info.subrun(), info.event()))

        # Process the run information
        self.process_run_info()

        # Process the entry list
        self.process_entry_list(
            n_entry,
            n_skip,
            entry_list,
            skip_entry_list,
            run_event_list,
            skip_run_event_list,
            allow_missing,
        )

    def get(self, idx: int) -> Dict[str, Any]:
        """Returns a specific entry in the file.

        Parameters
        ----------
        idx : int
            Integer entry ID to access

        Returns
        -------
        dict
            Dictionary which maps each data product key to an entry in the tree
        """
        # Get the appropriate entry index
        assert idx < len(self.entry_index)
        entry_idx = self.entry_index[idx]

        # If this is the first data loading, instantiate chains
        if not self.trees_ready:
            for key in self.trees:
                chain = ROOT.TChain(f"{key}_tree")  # pylint: disable=E1101
                for f in self.file_paths:
                    chain.AddFile(f)
                self.trees[key] = chain
            self.trees_ready = True

        # Move the entry pointer
        for tree in self.trees.values():
            tree.GetEntry(entry_idx)

        # Load the relevant data products
        data = {}
        for key, tree in self.trees.items():
            data[key] = getattr(tree, f"{key}_branch")

        return data

    @staticmethod
    def list_data(file_path: str) -> Dict[str, List[str]]:
        """Dumps top-level information about the contents of a LArCV root file.

        Parameters
        ----------
        file_path : str
            Path to the file to scan

        Returns
        -------
        dict
            Dictionary which maps data types onto a list of keys
        """
        # Load up the file
        f = ROOT.TFile.Open(file_path, "r")  # pylint: disable=E1101

        # Loop over the list of keys
        data = {"sparse3d": [], "cluster3d": [], "particle": []}
        for k in f.GetListOfKeys():
            # The the key name
            name = k.GetName()

            # Only look at tree names
            if not name.endswith("_tree"):
                continue
            if len(name.split("_")) < 3:
                continue

            # Get the data type name, skip if not recognized
            key = name.split("_")[0]
            if key not in data:
                continue

            # Append this specific tree name
            data[key].append(name[: name.rfind("_")])

        return data



"""Contains reusable layers for various models. Layers are classified into three folders:

* `common` reusable by any models
* `cnn` reusable by CNN-based models
* `cluster` reusable by CNN-based clustering models
* `gnn` reusable by GNN-based models
"""
